"""Functions related to mapping parameter from model to parameter estimation
problem"""

import pandas as pd
import numpy as np
import libsbml
import numbers
import re
from . import lint
from . import core
from typing import List, Tuple, Dict, Union, Any
import logging

logger = logging.getLogger(__name__)

# Parameter mapping for condition
ParMappingDict = Dict[str, Union[str, numbers.Number]]
# Parameter mapping for combination of preequilibration and simulation
# condition
ParMappingDictTuple = Tuple[ParMappingDict, ParMappingDict]
# Same for scale mapping
ScaleMappingDict = Dict[str, str]
ScaleMappingDictTuple = Tuple[ScaleMappingDict, ScaleMappingDict]


def get_optimization_to_simulation_parameter_mapping(
        condition_df: pd.DataFrame,
        measurement_df: pd.DataFrame,
        parameter_df: pd.DataFrame = None,
        sbml_model: libsbml.Model = None,
        simulation_conditions=None,
        warn_unmapped: bool = True) -> List[ParMappingDictTuple]:
    """
    Create list of mapping dicts from PEtab-problem to SBML parameters.

    Parameters
    ----------
    condition_df, measurement_df, parameter_df:
        The dataframes in the PEtab format.

        parameter_df is optional if par_sim_ids is provided

    sbml_model:
        The sbml model with observables and noise specified according to the
        petab format. Optional if par_sim_ids is provided.

    simulation_conditions: pd.DataFrame
        Table of simulation conditions as created by
        `petab.get_simulation_conditions`.

    warn_unmapped:
        If True, log warning regarding unmapped parameters

    Returns
    -------
    The length of the returned array is n_conditions, each entry is a tuple of
    two dicts of length n_par_sim, listing the optimization parameters or
    constants to be mapped to the simulation parameters, first for
    preequilibration (empty if no preequilibration condition is specified),
    second for simulation. NaN is used where no mapping exists.
    """
    # Ensure inputs are okay
    perform_mapping_checks(measurement_df)

    if simulation_conditions is None:
        simulation_conditions = core.get_simulation_conditions(measurement_df)

    mapping = []
    for condition_ix, condition in simulation_conditions.iterrows():
        cur_measurement_df = core.get_rows_for_condition(
            measurement_df, condition)

        if 'preequilibrationConditionId' not in condition \
                or not isinstance(condition.preequilibrationConditionId, str) \
                or not condition.preequilibrationConditionId:
            preeq_map = {}
        else:
            preeq_map = get_parameter_mapping_for_condition(
                condition_id=condition.preequilibrationConditionId,
                is_preeq=True,
                cur_measurement_df=cur_measurement_df,
                condition_df=condition_df,
                parameter_df=parameter_df, sbml_model=sbml_model,
                warn_unmapped=warn_unmapped
            )

        sim_map = get_parameter_mapping_for_condition(
            condition_id=condition.simulationConditionId,
            is_preeq=False,
            cur_measurement_df=cur_measurement_df,
            condition_df=condition_df,
            parameter_df=parameter_df, sbml_model=sbml_model,
            warn_unmapped=warn_unmapped
        )
        mapping.append((preeq_map, sim_map),)

    return mapping


def get_parameter_mapping_for_condition(
        condition_id: str,
        is_preeq: bool,
        cur_measurement_df: pd.DataFrame,
        condition_df: pd.DataFrame,
        parameter_df: pd.DataFrame = None,
        sbml_model: libsbml.Model = None,
        warn_unmapped: bool = True) -> ParMappingDict:
    """
    Create dictionary of mappings from PEtab-problem to SBML parameters for the
    given condition.

    Parameters
    ----------
    condition_id: Condition ID for which to perform mapping

    is_preeq: If true, output parameters will not be mapped

    cur_measurement_df: Measurement sub-table for current condition

    condition_df, parameter_df:
        The dataframes in the PEtab format.
        parameter_df is optional if par_sim_ids is provided

    sbml_model:
        The sbml model with observables and noise specified according to the
        petab format. Optional if par_sim_ids is provided.

    warn_unmapped:
        If True, log warning regarding unmapped parameters

    Returns
    -------
    Dictionary of parameter IDs with mapped parameters IDs to be estimated or
    filled in values in case of non-estimated parameters. NaN is used where no
    mapping exists.
    """
    perform_mapping_checks(cur_measurement_df)

    par_sim_ids = core.get_model_parameters(sbml_model)

    # initialize mapping dict
    # for the case of matching simulation and optimization parameter vector
    mapping = {par: par for par in par_sim_ids}

    _apply_dynamic_parameter_overrides(mapping, condition_id, condition_df)

    if not is_preeq:
        _apply_output_parameter_overrides(mapping, cur_measurement_df)

    fill_in_nominal_values(mapping, parameter_df)

    # TODO fill in fixed parameters (#103)

    handle_missing_overrides(mapping, warn=warn_unmapped)
    return mapping


def _apply_output_parameter_overrides(
        mapping: ParMappingDict,
        cur_measurement_df: pd.DataFrame) -> None:
    """
    Apply output parameter overrides to the parameter mapping dict for a given
    condition as defined in the measurement table (observableParameter,
    noiseParameters).

    Arguments:
        mapping: parameter mapping dict
        cur_measurement_df:
            Subset of the measurement table for the current condition
    """
    for _, row in cur_measurement_df.iterrows():
        # we trust that the number of overrides matches (see above)
        overrides = core.split_parameter_replacement_list(
            row.observableParameters)
        _apply_overrides_for_observable(mapping, row.observableId,
                                        'observable', overrides)

        overrides = core.split_parameter_replacement_list(row.noiseParameters)
        _apply_overrides_for_observable(mapping, row.observableId, 'noise',
                                        overrides)


def _apply_overrides_for_observable(
        mapping: ParMappingDict,
        observable_id: str,
        override_type: str,
        overrides: list) -> None:
    """
    Apply parameter-overrides for observables and noises to mapping
    matrix.

    Arguments:
        mapping: mapping dict to which to apply overrides
        observable_id: observable ID
        override_type: 'observable' or 'noise'
        overrides: list of overrides for noise or observable parameters
    """
    for i, override in enumerate(overrides):
        overridee_id = f'{override_type}Parameter{i+1}_{observable_id}'
        mapping[overridee_id] = override


def _apply_dynamic_parameter_overrides(mapping: ParMappingDict,
                                       condition_id: str,
                                       condition_df: pd.DataFrame) -> None:
    """Apply dynamic parameter overrides from condition table (in-place).

    Arguments:
        mapping, par_sim_id_to_ix:
            see get_parameter_mapping_for_condition
        condition_df: PEtab condition and parameter table
    """
    for overridee_id in condition_df.columns:
        if overridee_id == 'conditionName':
            continue
        if condition_df[overridee_id].dtype != 'O':
            continue

        overrider_id = condition_df.loc[condition_id, overridee_id]
        mapping[overridee_id] = overrider_id


def fill_in_nominal_values(mapping: ParMappingDict,
                           parameter_df: pd.DataFrame) -> None:
    """Replace non-estimated parameters in mapping list for a given condition
    by nominalValues provided in parameter table.

    Arguments:
        mapping:
            mapping dict obtained from get_parameter_mapping_for_condition
        parameter_df:
            PEtab parameter table
    """

    if parameter_df is None:
        return
    if 'estimate' not in parameter_df:
        return

    overrides = {row.name: row.nominalValue for _, row
                 in parameter_df.iterrows() if row.estimate != 1}

    for par, overridee in mapping.items():
        if not isinstance(overridee, str):
            continue

        try:
            mapping[par] = overrides[overridee]
            # all overrides will be scaled to 'lin'
            if 'parameterScale' in parameter_df:
                scale = parameter_df.loc[overridee, 'parameterScale']
                if scale == 'log':
                    mapping[par] = np.exp(mapping[par])
                elif scale == 'log10':
                    mapping[par] = np.power(10, mapping[par])
        except KeyError:
            # parameter is to be estimated
            pass


def get_optimization_to_simulation_scale_mapping(
        parameter_df: pd.DataFrame,
        mapping_par_opt_to_par_sim: List[ParMappingDictTuple],
        measurement_df: pd.DataFrame,
        simulation_conditions: Union[dict, pd.DataFrame] = None
) -> List[ScaleMappingDictTuple]:
    """Get parameter scale mapping for all conditions"""
    mapping_scale_opt_to_scale_sim = []

    if simulation_conditions is None:
        simulation_conditions = core.get_simulation_conditions(measurement_df)

    # iterate over conditions
    for condition_ix, condition in simulation_conditions.iterrows():
        if 'preequilibrationConditionId' not in condition \
                or not isinstance(condition.preequilibrationConditionId, str) \
                or not condition.preequilibrationConditionId:
            preeq_map = {}
        else:
            preeq_map = get_scale_mapping_for_condition(
                parameter_df=parameter_df,
                mapping_par_opt_to_par_sim=mapping_par_opt_to_par_sim[
                    condition_ix][0]
            )

        sim_map = get_scale_mapping_for_condition(
            parameter_df=parameter_df,
            mapping_par_opt_to_par_sim=mapping_par_opt_to_par_sim[
                condition_ix][1]
        )

        # append to mapping
        mapping_scale_opt_to_scale_sim.append((preeq_map, sim_map),)

    return mapping_scale_opt_to_scale_sim


def get_scale_mapping_for_condition(
        parameter_df: pd.DataFrame,
        mapping_par_opt_to_par_sim: ParMappingDict) -> ScaleMappingDict:
    """Get parameter scale mapping for the given condition.

    Arguments:
        parameter_df: PEtab parameter table
        mapping_par_opt_to_par_sim:
            Mapping as obtained from get_parameter_mapping_for_condition
    """
    def get_scale(par_id_or_val):
        if isinstance(par_id_or_val, numbers.Number):
            # fixed value assignment
            return 'lin'
        else:
            # is par opt id, thus extract its scale
            try:
                return parameter_df.loc[par_id_or_val, 'parameterScale']
            except KeyError:
                # This is a condition-table parameter which is not
                # present in the parameter table. Those are assumed to be
                # 'lin'
                return 'lin'

    return {par: get_scale(val)
            for par, val in mapping_par_opt_to_par_sim.items()}


def perform_mapping_checks(measurement_df: pd.DataFrame) -> None:
    if lint.measurement_table_has_timepoint_specific_mappings(measurement_df):
        # we could allow that for floats, since they don't matter in this
        # function and would be simply ignored
        raise ValueError(
            "Timepoint-specific parameter overrides currently unsupported.")


def handle_missing_overrides(mapping_par_opt_to_par_sim: ParMappingDict,
                             warn: bool = True,
                             condition_id: str = None) -> None:
    """
    Find all observable parameters and noise parameters that were not mapped
    and set their mapping to np.nan.

    Assumes that parameters matching "(noise|observable)Parameter[0-9]+_" were
    all supposed to be overwritten.

    Parameters:
    -----------
    mapping_par_opt_to_par_sim:
        Output of get_parameter_mapping_for_condition
    warn:
        If True, log warning regarding unmapped parameters
    """
    _missed_vals = []
    rex = re.compile("^(noise|observable)Parameter[0-9]+_")
    for key, val in mapping_par_opt_to_par_sim.items():
        try:
            matches = rex.match(val)
        except TypeError:
            continue

        if matches:
            mapping_par_opt_to_par_sim[key] = np.nan
            _missed_vals.append(key)

    if len(_missed_vals) and warn:
        logger.warning(f"Could not map the following overrides for condition "
                       f"{condition_id}: "
                       f"{_missed_vals}. Usually, this is just due to missing "
                       f"data points.")


def merge_preeq_and_sim_pars_condition(
        condition_map_preeq: ParMappingDict,
        condition_map_sim: ParMappingDict,
        condition_scale_map_preeq: ScaleMappingDict,
        condition_scale_map_sim: ScaleMappingDict,
        condition: Any) -> None:
    """Merge preequilibration and simulation parameters and scales while
    checking for compatibility.

    This function is meant for the case where we cannot have different
    parameters (and scales) for preequilibration and simulation. Therefore,
    merge both and ensure matching scales and parameters.
    `condition_map_sim` and `condition_scale_map_sim` will ne modified in
    place.

    Arguments:
        condition_map_preeq, condition_map_sim:
            Parameter mapping as obtained from
            `get_parameter_mapping_for_condition`
        condition_scale_map_preeq, condition_scale_map_sim:
            Parameter scale mapping as obtained from
            `get_get_scale_mapping_for_condition`
        condition: Condition identifier for more informative error messages
    """
    if not condition_map_preeq:
        # nothing to do
        return

    for idx, (par_preeq, par_sim, scale_preeq, scale_sim) \
            in enumerate(zip(condition_map_preeq,
                             condition_map_sim,
                             condition_scale_map_preeq,
                             condition_scale_map_sim)):
        if par_preeq != par_sim \
                and not (np.isnan(par_sim) and np.isnan(par_preeq)):
            # both identical or both nan is okay
            if np.isnan(par_sim):
                # unmapped for simulation
                par_sim[idx] = par_preeq
            elif np.isnan(par_preeq):
                # unmapped for preeq is okay
                pass
            else:
                raise ('Cannot handle different values for dynamic '
                       f'parameters: for condition {condition} '
                       f'parameter {idx} is {par_preeq} for preeq '
                       f'and {par_sim} for simulation.')
        if scale_preeq != scale_sim:
            # both identical is okay
            if np.isnan(par_sim):
                # unmapped for simulation
                scale_sim[idx] = scale_preeq
            elif np.isnan(par_preeq):
                # unmapped for preeq is okay
                pass
            else:
                raise ('Cannot handle different parameter scales '
                       f'parameters: for condition {condition} '
                       f'scale for parameter {idx} is '
                       f'{scale_preeq} for preeq '
                       f'and {scale_sim} for simulation.')



import pandas as pd
import numpy as np
import numbers


def generate_experiment_id(measurement_data):
    '''
    automatically generate the experimentId:
    every row which shares the observableParameters (if provided),
    noiseParameters and observableTransformation gets
    the same experimentId, for the assignment all numeric values in the
    noiseParameters are considered to be the same

    Parameters:
    ----------

    measurement_data: pandas data frame for measurement data

    Return:
    ----------

    measurement_data: pandas data frame for measurement data with column for
    experimentId added

    '''
    observable_parameters = np.array(measurement_data.observableParameters)
    noise_parameters = np.array(measurement_data.noiseParameters)
    observable_transformation = np.array(
        measurement_data.observableTransformation)

    # check if there is empty values in the observableParameter column, should
    # not occur!
    if np.any(observable_parameters == 'empty'):
        raise AssertionError("NOOOOOOO! observable_parameter must have an "
                             "entry!")

    # noiseParameters should always has an entry
    if np.any(noise_parameters == 'empty'):
        raise AssertionError("NOOOOOOO! noise_parameters must have an entry!")

    # help functions to check for numeric values and nans in arrays
    isnumeric = np.vectorize(isinstance)
    isnan_vectorized = np.vectorize(np.isnan)

    # assign 'empty' to missing observableParameters
    tmp_observable_parameters = observable_parameters
    tmp_ind = np.where(isnumeric(observable_parameters, numbers.Number))
    if tmp_ind[0].size == 0:
        tmp_observable_parameters == observable_parameters
    else:
        tmp_observable_parameters[tmp_ind[0][
            np.where(isnan_vectorized(observable_parameters[tmp_ind[0]]))]
        ] = 'empty'

    # all numeric values in noiseParameters are temporarily set to 0 to be
    # treated the same
    tmp_noise_parameters = noise_parameters
    tmp_noise_parameters[
        np.where(isnumeric(noise_parameters, numbers.Number))
    ] = 0

    # add new column for experimentId
    measurement_data = measurement_data.assign(
        experimentId=pd.Series(np.zeros(len(measurement_data)))
    )

    ind_no_exp_id = np.array(range(len(measurement_data.experimentId)))
    count = 1
    # assign experiment ID when data share the same observable parameters,
    # noise parameters and observable transformation
    while ind_no_exp_id.size > 0:
        ind_exp_id = np.where(
            (observable_parameters == observable_parameters[ind_no_exp_id[
                0]]) *
            (tmp_noise_parameters == tmp_noise_parameters[ind_no_exp_id[
                0]]) *
            (observable_transformation ==
             observable_transformation[ind_no_exp_id[0]]))
        for ind in ind_exp_id[0]:
            measurement_data.loc[ind, 'experimentId'] = 'experiment_' + str(
                count)

        # extract measurements with no assigned experimentId
        ind_no_exp_id = np.where(measurement_data.experimentId == 0)[0]

        # if it does not decrease there might be a missing value in the
        # observableTransformation or noiseParameter
        print(str(ind_no_exp_id.size) + ' measurements left to be assigned')
        count = count + 1

    print(str(count - 1) + ' experimentIds added.')
    return measurement_data



from .core import *  # noqa: F403, F401
from .lint import *  # noqa: F403, F401
from .sbml import *  # noqa: F403, F401
from .parameter_mapping import *  # noqa: F403, F401
from .generate_experiment_id import generate_experiment_id  # noqa: F403, F401
from .version import __version__  # noqa: F401



"""Integrity checks and tests for specific features used"""

from . import core
from . import sbml
import numpy as np
import numbers
import re
import copy
import logging
import libsbml
import pandas as pd

logger = logging.getLogger(__name__)


def _check_df(df, req_cols, name):
    cols_set = df.columns.values
    missing_cols = set(req_cols) - set(cols_set)
    if missing_cols:
        raise AssertionError(
            f"Dataframe {name} requires the columns {missing_cols}.")


def assert_no_leading_trailing_whitespace(names_list, name):
    r = re.compile(r'(?:^\s)|(?:\s$)')
    for i, x in enumerate(names_list):
        if isinstance(x, str) and r.search(x):
            raise AssertionError(f"Whitespace around {name}[{i}] = '{x}'.")


def check_condition_df(df):
    req_cols = []
    _check_df(df, req_cols, "condition")

    if not df.index.name == 'conditionId':
        raise AssertionError(
            f"Condition table has wrong index {df.index.name}."
            "expected 'conditionId'.")

    assert_no_leading_trailing_whitespace(df.index.values, "conditionId")

    for column_name in req_cols:
        if not np.issubdtype(df[column_name].dtype, np.number):
            assert_no_leading_trailing_whitespace(
                df[column_name].values, column_name)


def check_measurement_df(df):
    req_cols = [
        "observableId", "preequilibrationConditionId", "simulationConditionId",
        "measurement", "time", "observableParameters", "noiseParameters"
    ]

    for column_name in req_cols:
        if not np.issubdtype(df[column_name].dtype, np.number):
            assert_no_leading_trailing_whitespace(
                df[column_name].values, column_name)

    _check_df(df, req_cols, "measurement")


def check_parameter_df(df):
    req_cols = [
        "parameterName", "parameterScale",
        "lowerBound", "upperBound", "nominalValue", "estimate"
    ]
    _check_df(df, req_cols, "parameter")

    if not df.index.name == 'parameterId':
        raise AssertionError(
            f"Parameter table has wrong index {df.index.name}."
            "expected 'parameterId'.")

    assert_no_leading_trailing_whitespace(df.index.values, "parameterId")

    for column_name in req_cols:
        if not np.issubdtype(df[column_name].dtype, np.number):
            assert_no_leading_trailing_whitespace(
                df[column_name].values, column_name)

    assert_parameter_id_is_string(df)
    assert_parameter_scale_is_valid(df)
    assert_parameter_bounds_are_numeric(df)
    assert_parameter_estimate_is_boolean(df)
    assert_parameter_id_is_unique(df)
    check_parameter_bounds(df)


def assert_measured_observables_present_in_model(measurement_df, sbml_model):
    """Check if all observables in measurement files have been specified in
    the model"""

    measurement_observables = [f'observable_{x}' for x in
                               measurement_df.observableId.values]

    model_observables = core.get_observables(sbml_model)
    undefined_observables = set(measurement_observables) - set(
        model_observables.keys())

    if len(undefined_observables):
        raise AssertionError(
            f"Unknown observables in measurement file: "
            f"{undefined_observables}.")


def condition_table_is_parameter_free(condition_df):
    """Check if all entries in the condition table are numeric
    (no parameter IDs)"""

    constant_parameters = list(
        set(condition_df.columns.values.tolist()) - {'conditionId',
                                                     'conditionName'})

    for column in constant_parameters:
        if np.any(np.invert(condition_df.loc[:, column].apply(
                isinstance, args=(numbers.Number,)))):
            return False
    return True


def assert_parameter_id_is_string(parameter_df):
    """
    Check if all entries in the parameterId column of the parameter table
    are string and not empty.
    """

    for parameterId in parameter_df:
        if isinstance(parameterId, str):
            if parameterId[0].isdigit():
                raise AssertionError('parameterId ' + parameterId
                                     + ' starts with integer')
        else:
            raise AssertionError('Empty parameterId found')


def assert_parameter_id_is_unique(parameter_df):
    """
    Check if the parameterId column of the parameter table is unique.
    """
    if len(parameter_df.index) != len(set(parameter_df.index)):
        raise AssertionError(
            'parameterId column in parameter table is not unique')


def assert_parameter_scale_is_valid(parameter_df):
    """
    Check if all entries in the parameterScale column of the parameter table
    are 'lin' for linear, 'log' for natural logarithm or 'log10' for base 10
    logarithm.
    """

    for parameterScale in parameter_df['parameterScale']:
        if parameterScale not in ['lin', 'log', 'log10']:
            raise AssertionError(
                'Expected "lin", "log" or "log10" but got "' +
                parameterScale + '"')


def assert_parameter_bounds_are_numeric(parameter_df):
    """
    Check if all entries in the lowerBound and upperBound columns of the
    parameter table are numeric.
    """
    parameter_df["lowerBound"].apply(float).all()
    parameter_df["upperBound"].apply(float).all()


def check_parameter_bounds(parameter_df):
    """
    Check if all entries in the lowerBound are smaller than upperBound column
    in the parameter table.
    """
    for element in range(len(parameter_df['lowerBound'])):
        if int(parameter_df['estimate'][element]):
            if not parameter_df['lowerBound'][element] \
                    <= parameter_df['upperBound'][element]:
                raise AssertionError(
                    f"lowerbound greater than upperBound for parameterId "
                    f"{parameter_df.index[element]}.")


def assert_parameter_estimate_is_boolean(parameter_df):
    """
    Check if all entries in the estimate column of the parameter table are
    0 or 1.
    """
    for estimate in parameter_df['estimate']:
        if not int(estimate) in [True, False]:
            raise AssertionError(
                f"Expected 0 or 1 but got {estimate} in estimate column.")


def measurement_table_has_timepoint_specific_mappings(measurement_df):
    """
    Are there time-point or replicate specific parameter assignments in the
    measurement table.
    """
    # since we edit it, copy it first
    measurement_df = copy.deepcopy(measurement_df)

    measurement_df.loc[
        measurement_df.noiseParameters.apply(isinstance, args=(
            numbers.Number,)), 'noiseParameters'] = np.nan

    grouping_cols = core.get_notnull_columns(
        measurement_df,
        ['observableId',
         'simulationConditionId',
         'preequilibrationConditionId',
         'observableParameters',
         'noiseParameters'
         ])
    grouped_df = measurement_df.groupby(grouping_cols).size().reset_index()

    grouping_cols = core.get_notnull_columns(
        grouped_df,
        ['observableId',
         'simulationConditionId',
         'preequilibrationConditionId'])
    grouped_df2 = grouped_df.groupby(grouping_cols).size().reset_index()

    if len(grouped_df.index) != len(grouped_df2.index):
        logger.warning(
            "Measurement table has timepoint specific mappings:\n"
            + str(grouped_df))
        return True
    return False


def measurement_table_has_observable_parameter_numeric_overrides(
        measurement_df):
    """Are there any numbers to override observable parameters?"""

    for i, row in measurement_df.iterrows():
        for override in core.split_parameter_replacement_list(
                row.observableParameters):
            if isinstance(override, numbers.Number):
                return True
    return False


def assert_noise_distributions_valid(measurement_df):
    """
    Check whether there are not multiple noise distributions for an
    observable, and that the names are correct.
    """
    df = measurement_df.copy()

    # insert optional columns into copied df

    if 'observableTransformation' not in df:
        df['observableTransformation'] = ''
    if 'noiseDistribution' not in df:
        df['noiseDistribution'] = ''

    # check for valid values

    for trafo in df['observableTransformation']:
        if trafo not in ['', 'lin', 'log', 'log10'] \
                and not (isinstance(trafo, numbers.Number)
                         and np.isnan(trafo)):
            raise ValueError(
                f"Unrecognized observable transformation in measurement "
                f"file: {trafo}.")
    for distr in df['noiseDistribution']:
        if distr not in ['', 'normal', 'laplace'] \
                and not (isinstance(distr, numbers.Number)
                         and np.isnan(distr)):
            raise ValueError(
                f"Unrecognized noise distribution in measurement "
                f"file: {distr}.")

    # check for unique values per observable

    distrs = df.groupby(['observableId']).size().reset_index()

    distrs_check = df.groupby(
        ['observableId', 'observableTransformation', 'noiseDistribution'])

    if len(distrs) != len(distrs_check):
        raise AssertionError(
            f"The noiseDistribution for an observable in the measurement "
            f"file is not unique: \n{distrs_check}")


def assert_overrides_match_parameter_count(measurement_df, observables, noise):
    """Ensure that number of parameters in the observable definition matches
    the number of overrides in `measurement_df`

    Arguments:
        :param measurement_df:
        :param observables: dict: obsId => {obsFormula}
        :param noise: dict: obsId => {obsFormula}
    """

    # sympify only once and save number of parameters
    observable_parameters_count = {oid[len('observable_'):]:
                                   len(core.get_placeholders(
                                       value['formula'],
                                       oid[len('observable_'):],
                                       'observable'))
                                   for oid, value in observables.items()}
    noise_parameters_count = {
        oid[len('observable_'):]: len(core.get_placeholders(
            value, oid[len('observable_'):], 'noise'))
        for oid, value in noise.items()
    }

    for _, row in measurement_df.iterrows():
        # check observable parameters
        try:
            expected = observable_parameters_count[row.observableId]
        except KeyError:
            raise ValueError(
                f"Observable {row.observableId} used in measurement table "
                f"but not defined in model {observables.keys()}.")
        actual = len(
            core.split_parameter_replacement_list(row.observableParameters))
        # No overrides are also allowed
        if not (actual == 0 or actual == expected):
            raise AssertionError(
                f'Mismatch of observable parameter overrides for '
                f'{observables[f"observable_{row.observableId}"]} '
                f'in:\n{row}\n'
                f'Expected 0 or {expected} but got {actual}')

        # check noise parameters
        replacements = core.split_parameter_replacement_list(
            row.noiseParameters)
        try:
            expected = noise_parameters_count[row.observableId]

            # No overrides are also allowed
            if not (len(replacements) == 0 or len(replacements) == expected):
                raise AssertionError(
                    f'Mismatch of noise parameter overrides in:\n{row}\n'
                    f'Expected 0 or {expected} but got {actual}')
        except KeyError:
            # no overrides defined, but a numerical sigma can be provided
            # anyways
            if not len(replacements) == 1 \
                    or not isinstance(replacements[0], numbers.Number):
                raise AssertionError(
                    f'No placeholders have been specified in the noise model '
                    f'SBML AssigmentRule for: '
                    f'\n{row}\n'
                    f'But parameter name or multiple overrides were specified '
                    'in the noiseParameters column.')


def lint_problem(problem: 'core.Problem'):
    """Run PEtab validation on problem

    Arguments:
        problem: PEtab problem to check

    Returns:
        True is errors occurred, False otherwise
    """
    errors_occurred = False

    # Run checks on individual files
    if problem.sbml_model is not None:
        logger.info("Checking SBML model...")
        errors_occurred |= not sbml.is_sbml_consistent(
            problem.sbml_model.getSBMLDocument())
        sbml.log_sbml_errors(problem.sbml_model.getSBMLDocument())
    else:
        logger.warning("SBML model not available. Skipping.")

    if problem.measurement_df is not None:
        logger.info("Checking measurement table...")
        try:
            check_measurement_df(problem.measurement_df)
        except AssertionError as e:
            logger.error(e)
            errors_occurred = True
    else:
        logger.warning("Measurement table not available. Skipping.")

    if problem.condition_df is not None:
        logger.info("Checking condition table...")
        try:
            check_condition_df(problem.condition_df)
        except AssertionError as e:
            logger.error(e)
            errors_occurred = True
    else:
        logger.warning("Condition table not available. Skipping.")

    if problem.parameter_df is not None:
        logger.info("Checking parameter table...")
        try:
            check_parameter_df(problem.parameter_df)
        except AssertionError as e:
            logger.error(e)
            errors_occurred = True
    else:
        logger.warning("Parameter table not available. Skipping.")

    if problem.measurement_df is not None and problem.sbml_model is not None \
            and not errors_occurred:
        try:
            assert_measured_observables_present_in_model(
                problem.measurement_df, problem.sbml_model)
            assert_overrides_match_parameter_count(
                problem.measurement_df,
                core.get_observables(problem.sbml_model, remove=False),
                core.get_sigmas(problem.sbml_model, remove=False)
            )
        except AssertionError as e:
            logger.error(e)
            errors_occurred = True

    if problem.sbml_model is not None and problem.condition_df is not None \
            and problem.parameter_df is not None:
        try:
            assert_model_parameters_in_condition_or_parameter_table(
                problem.sbml_model,
                problem.condition_df,
                problem.parameter_df
            )
        except AssertionError as e:
            logger.error(e)
            errors_occurred = True

    if errors_occurred:
        logger.error('Not OK')
    elif problem.measurement_df is None or problem.condition_df is None \
            or problem.sbml_model is None or problem.parameter_df is None:
        logger.warning('Not all files of the PEtab problem definition could '
                       'be checked.')
    else:
        logger.info('OK')

    return errors_occurred


def assert_model_parameters_in_condition_or_parameter_table(
        sbml_model: libsbml.Model,
        condition_df: pd.DataFrame,
        parameter_df: pd.DataFrame):
    """Model parameters that are targets of AssignmentRule must not be present
    in parameter table or in condition table columns. Other parameters must
    only be present in either in parameter table or condition table columns.
    Check that."""

    for parameter in sbml_model.getListOfParameters():
        parameter_id = parameter.getId()

        if parameter_id.startswith('observableParameter'):
            continue
        if parameter_id.startswith('noiseParameter'):
            continue

        is_assignee = \
            sbml_model.getAssignmentRuleByVariable(parameter_id) is not None
        in_parameter_df = parameter_id in parameter_df.index
        in_condition_df = parameter_id in condition_df.columns

        if is_assignee and (in_parameter_df or in_condition_df):
            raise AssertionError(f"Model parameter '{parameter_id}' is target "
                                 "of AssignmentRule, and thus, must not be "
                                 "present in condition table or in parameter "
                                 "table.")

        if in_parameter_df and in_condition_df:
            raise AssertionError(f"Model parameter '{parameter_id}' present "
                                 "in both condition table and parameter "
                                 "table.")
